/* GStreamer
 * Copyright (C) 2010 Jan Schmidt <thaytan@noraisin.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>

#include <gst/gst.h>
#include <gst/pbutils/missing-plugins.h>

#include "gstinteractbin.h"
#include "gstinteractsrc.h"

GST_DEBUG_CATEGORY_EXTERN (gst_interact_debug);
#define GST_CAT_DEFAULT gst_interact_debug

#define GST_IBIN_LOCK(d) g_mutex_lock((d)->main_lock)
#define GST_IBIN_UNLOCK(d) g_mutex_unlock((d)->main_lock)

#define GST_IBIN_PREROLL_LOCK(d) g_mutex_lock((d)->preroll_lock)
#define GST_IBIN_PREROLL_UNLOCK(d) g_mutex_unlock((d)->preroll_lock)

#define DEFAULT_LOCATION ""
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  ARG_0,
  ARG_LOCATION
};

static GstStaticPadTemplate video_src_template =
GST_STATIC_PAD_TEMPLATE ("video",
    GST_PAD_SRC,
    GST_PAD_SOMETIMES,
    GST_STATIC_CAPS ("video/x-raw-rgb")
    );

static GstStaticPadTemplate audio_src_template =
    GST_STATIC_PAD_TEMPLATE ("audio",
    GST_PAD_SRC,
    GST_PAD_SOMETIMES,
    GST_STATIC_CAPS ("audio/x-raw-int; audio/x-raw-float")
    );

static GstStaticPadTemplate subpicture_src_template =
GST_STATIC_PAD_TEMPLATE ("subpicture",
    GST_PAD_SRC,
    GST_PAD_SOMETIMES,
    GST_STATIC_CAPS ("video/x-dvd-subpicture")
    );

static void gst_interactbin_do_init (GType gst_interactbin_type);
static void gst_interactbin_finalize (GObject * object);
static void gst_interactbin_uri_handler_init (gpointer g_iface, gpointer iface_data);

GST_BOILERPLATE_FULL (GstInteractBin, gst_interactbin, GstBin,
    GST_TYPE_BIN, gst_interactbin_do_init);

static void gst_interactbin_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_interactbin_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static GstStateChangeReturn gst_interactbin_change_state (GstElement * element,
    GstStateChange transition);
static GstPad *connect_thru_mq (GstInteractBin *ibin, GstPad * pad);

static void
gst_interactbin_base_init (gpointer gclass)
{

  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&video_src_template));
/*
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&audio_src_template));
*/
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&subpicture_src_template));

  gst_element_class_set_details_simple (element_class,
      "Interactive SVG slide menu system",
      "Generic/Bin/Player",
      "Demonstration element for GstNavigation event handling",
      "Jan Schmidt <thaytan@noraisin.net>");

  element_class->change_state = GST_DEBUG_FUNCPTR (gst_interactbin_change_state);
}

static void
gst_interactbin_class_init (GstInteractBinClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->finalize = gst_interactbin_finalize;
  gobject_class->set_property = gst_interactbin_set_property;
  gobject_class->get_property = gst_interactbin_get_property;

  g_object_class_install_property (gobject_class, ARG_LOCATION,
      g_param_spec_string ("location", "Location",
          "Location of the slides.txt file to present",
          NULL, G_PARAM_READWRITE));
}

static void
gst_interactbin_do_init (GType gst_interactbin_type)
{
  static const GInterfaceInfo urihandler_info = {
    gst_interactbin_uri_handler_init,
    NULL,
    NULL
  };

  g_type_add_interface_static (gst_interactbin_type, GST_TYPE_URI_HANDLER,
      &urihandler_info);
}

static void
gst_interactbin_init (GstInteractBin * ibin, GstInteractBinClass * gclass)
{
  ibin->main_lock = g_mutex_new ();
  ibin->preroll_lock = g_mutex_new ();

  ibin->audio_added = TRUE;
}

static void
gst_interactbin_finalize (GObject * object)
{
  GstInteractBin *ibin = GST_INTERACTBIN (object);

  g_mutex_free (ibin->main_lock);
  g_mutex_free (ibin->preroll_lock);
  g_free (ibin->last_uri);
  g_free (ibin->location);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

/* URI interface */
static GstURIType
gst_interactbin_uri_get_type (void)
{
  return GST_URI_SRC;
}

static gchar **
gst_interactbin_uri_get_protocols (void)
{
  static gchar *protocols[] = { (char *) "interact", NULL };

  return protocols;
}

static const gchar *
gst_interactbin_uri_get_uri (GstURIHandler * handler)
{
  GstInteractBin *ibin = GST_INTERACTBIN (handler);

  GST_IBIN_LOCK (ibin);
  g_free (ibin->last_uri);
  if (ibin->location)
    ibin->last_uri = g_strdup_printf ("interact://%s", ibin->location);
  else
    ibin->last_uri = g_strdup ("interact://");
  GST_IBIN_UNLOCK (ibin);

  return ibin->last_uri;
}

static gboolean
gst_interactbin_uri_set_uri (GstURIHandler * handler, const gchar * uri)
{
  GstInteractBin *ibin = GST_INTERACTBIN (handler);
  gboolean ret;
  gchar *protocol, *location;

  protocol = gst_uri_get_protocol (uri);

  ret = (protocol && !strcmp (protocol, "interact")) ? TRUE : FALSE;

  g_free (protocol);
  protocol = NULL;

  if (!ret)
    return ret;

  location = gst_uri_get_location (uri);
  if (!location)
    return ret;

  /*
   * URI structure: interact:///path/to/slides
   */
  if (g_str_has_prefix (uri, "interact://")) {
    g_free (ibin->location);
    if (strlen (uri) > 11)
      ibin->location = g_strdup (uri + 11);
    else
      ibin->location = g_strdup (DEFAULT_LOCATION);
  }

  g_free (location);

  return ret;
}

static void
gst_interactbin_uri_handler_init (gpointer g_iface, gpointer iface_data)
{
  GstURIHandlerInterface *iface = (GstURIHandlerInterface *) g_iface;

  iface->get_type = gst_interactbin_uri_get_type;
  iface->get_protocols = gst_interactbin_uri_get_protocols;
  iface->get_uri = gst_interactbin_uri_get_uri;
  iface->set_uri = gst_interactbin_uri_set_uri;
}

static gboolean
try_create_piece (GstInteractBin * ibin, gint index,
    const gchar * factory, GType type, const gchar * name, const gchar * descr)
{
  GstElement *e;

  GST_IBIN_LOCK (ibin);
  if (ibin->pieces[index] != NULL) {
    GST_IBIN_UNLOCK (ibin);
    return TRUE;                /* Already exists */
  }
  GST_IBIN_UNLOCK (ibin);

  if (factory != NULL) {
    e = gst_element_factory_make (factory, name);
  } else {
    if (name)
      e = g_object_new (type, "name", name, NULL);
    else
      e = g_object_new (type, NULL);
  }
  if (e == NULL)
    goto create_failed;

  if (!gst_bin_add (GST_BIN (ibin), e))
    goto add_failed;

  GST_DEBUG_OBJECT (ibin, "Added %s element: %" GST_PTR_FORMAT, descr, e);

  GST_IBIN_LOCK (ibin);
  ibin->pieces[index] = e;
  GST_IBIN_UNLOCK (ibin);

  return TRUE;
create_failed:
  gst_element_post_message (GST_ELEMENT_CAST (ibin),
      gst_missing_element_message_new (GST_ELEMENT_CAST (ibin), factory));
  GST_ELEMENT_ERROR (ibin, CORE, MISSING_PLUGIN, (NULL),
      ("Could not create %s element '%s'", descr, factory));
  return FALSE;
add_failed:
  gst_object_unref (e);
  GST_ELEMENT_ERROR (ibin, CORE, FAILED, (NULL),
      ("Could not add %s element to bin", descr));
  return FALSE;
}

typedef struct
{
  GstInteractBin *ibin;
  GstPad *pad;
} GstInteractBinPadBlockCtx;

static void ibin_pad_blocked_cb (GstPad * pad, gboolean blocked,
    GstInteractBinPadBlockCtx * ctx);
static void
_pad_block_destroy_notify (GstInteractBinPadBlockCtx * ctx)
{
  gst_object_unref (ctx->ibin);
  gst_object_unref (ctx->pad);
  g_slice_free (GstInteractBinPadBlockCtx, ctx);
}

static gboolean
create_elements (GstInteractBin * ibin)
{
  GstPadTemplate *src_templ = NULL;
  GstPad *src = NULL;
  GstPad *mq_src = NULL;
  GstPad *sink = NULL;
  GstInteractBinPadBlockCtx *bctx = NULL;

  if (!try_create_piece (ibin, GSTI_ELEM_SOURCE, NULL,
          GST_TYPE_INTERACT_SRC, "interactsrc", "source")) {
    return FALSE;
  }

  /* FIXME: Locking */
  if (ibin->location) {
    g_object_set (G_OBJECT (ibin->pieces[GSTI_ELEM_SOURCE]),
        "location", ibin->location, NULL);
  }

  if (!try_create_piece (ibin, GSTI_ELEM_MQUEUE, "multiqueue", 0, "mq",
          "multiqueue"))
    return FALSE;

  g_object_set (ibin->pieces[GSTI_ELEM_MQUEUE],
      "max-size-time", (7 * GST_SECOND / 10), "max-size-bytes", 0,
      "max-size-buffers", 0, NULL);

  src = gst_element_get_static_pad (ibin->pieces[GSTI_ELEM_SOURCE],
            "src");
  mq_src = connect_thru_mq (ibin, src);
  gst_object_unref (src); src = NULL;

  src_templ = gst_static_pad_template_get (&video_src_template);
  ibin->video_pad = gst_ghost_pad_new_from_template ("video", mq_src, src_templ);
  gst_object_unref (src_templ);
  if (ibin->video_pad == NULL)
    goto failed_video_ghost;
  gst_pad_set_active (ibin->video_pad, TRUE);
  bctx = g_slice_new (GstInteractBinPadBlockCtx);
  bctx->ibin = gst_object_ref (ibin);
  bctx->pad = gst_object_ref (ibin->video_pad);
  gst_pad_set_blocked_async_full (mq_src, TRUE,
      (GstPadBlockCallback) ibin_pad_blocked_cb, bctx, (GDestroyNotify)
      _pad_block_destroy_notify);

  gst_object_unref (mq_src); mq_src = NULL;

  /* Add a single standalone queue to hold a single buffer of SPU data */
  if (!try_create_piece (ibin, GSTI_ELEM_SPUQ, "queue", 0, "spu_q",
          "subpicture decoder buffer"))
    return FALSE;
  g_object_set (ibin->pieces[GSTI_ELEM_SPUQ],
      "max-size-time", G_GUINT64_CONSTANT (0), "max-size-bytes", 0,
      "max-size-buffers", 1, NULL);

  src = gst_element_get_static_pad (ibin->pieces[GSTI_ELEM_SOURCE],
            "subpicture");
  sink = gst_element_get_static_pad (ibin->pieces[GSTI_ELEM_SPUQ], "sink");
  if (src == NULL || sink == NULL)
    goto failed_spuq_connect;
  if (GST_PAD_LINK_FAILED (gst_pad_link (src, sink)))
    goto failed_spuq_connect;
  gst_object_unref (src);
  gst_object_unref (sink);
  src = sink = NULL;

  src = gst_element_get_static_pad (ibin->pieces[GSTI_ELEM_SPUQ], "src");
  if (src == NULL)
    goto failed_spu_ghost;
  src_templ = gst_static_pad_template_get (&subpicture_src_template);
  ibin->subpicture_pad =
      gst_ghost_pad_new_from_template ("subpicture", src, src_templ);
  gst_object_unref (src_templ);
  if (ibin->subpicture_pad == NULL)
    goto failed_spu_ghost;
  gst_pad_set_active (ibin->subpicture_pad, TRUE);
  bctx = g_slice_new (GstInteractBinPadBlockCtx);
  bctx->ibin = gst_object_ref (ibin);
  bctx->pad = gst_object_ref (ibin->subpicture_pad);
  gst_pad_set_blocked_async_full (src, TRUE,
      (GstPadBlockCallback) ibin_pad_blocked_cb, bctx, (GDestroyNotify)
      _pad_block_destroy_notify);
  gst_object_unref (src);
  src = NULL;

  return TRUE;

failed_spuq_connect:
  GST_ELEMENT_ERROR (ibin, CORE, FAILED, (NULL),
      ("Could not connect source to subpicture buffer element"));
  goto error_out;
failed_spu_ghost:
  GST_ELEMENT_ERROR (ibin, CORE, FAILED, (NULL),
      ("Could not ghost SPU output pad"));
  goto error_out;
failed_video_ghost:
  GST_ELEMENT_ERROR (ibin, CORE, FAILED, (NULL),
      ("Could not ghost video output pad"));
  goto error_out;
error_out:
  if (src != NULL)
    gst_object_unref (src);
  if (sink != NULL)
    gst_object_unref (sink);
  return FALSE;
}

static GstPad *
connect_thru_mq (GstInteractBin *ibin, GstPad * pad)
{
  GstPad *mq_sink;
  GstPad *mq_src;
  gchar *tmp, *sinkname, *srcname;

  /* Request a pad from multiqueue, then connect this one, then
   * discover the corresponding output pad and return it */
  mq_sink = gst_element_get_request_pad (ibin->pieces[GSTI_ELEM_MQUEUE],
      "sink%d");
  if (mq_sink == NULL)
    return FALSE;
  ibin->mq_req_pads = g_list_prepend (ibin->mq_req_pads, mq_sink);

  if (gst_pad_link (pad, mq_sink) != GST_PAD_LINK_OK)
    return FALSE;

  sinkname = gst_pad_get_name (mq_sink);
  tmp = sinkname + 4;
  srcname = g_strdup_printf ("src%s", tmp);

  mq_src = gst_element_get_static_pad (ibin->pieces[GSTI_ELEM_MQUEUE],
      srcname);

  g_free (sinkname);
  g_free (srcname);

  return mq_src;
}

static void
remove_elements (GstInteractBin * ibin)
{
  gint i;
  GList *tmp;

  if (ibin->pieces[GSTI_ELEM_MQUEUE] != NULL) {
    for (tmp = ibin->mq_req_pads; tmp; tmp = g_list_next (tmp)) {
      gst_element_release_request_pad (ibin->pieces[GSTI_ELEM_MQUEUE],
          GST_PAD (tmp->data));
    }
  }
  g_list_free (ibin->mq_req_pads);
  ibin->mq_req_pads = NULL;

  for (i = 0; i < GSTI_ELEM_LAST; i++) {
    GST_IBIN_LOCK (ibin);
    if (ibin->pieces[i] != NULL) {
      GstElement *piece = ibin->pieces[i];

      ibin->pieces[i] = NULL;
      GST_IBIN_UNLOCK (ibin);

      gst_element_set_state (piece, GST_STATE_NULL);
      gst_bin_remove (GST_BIN (ibin), piece);
    } else
      GST_IBIN_UNLOCK (ibin);
  }
  if (ibin->video_pad) {
    if (ibin->video_added)
      gst_element_remove_pad (GST_ELEMENT (ibin), ibin->video_pad);
    else
      gst_object_unref (ibin->video_pad);
  }
  if (ibin->audio_pad) {
    if (ibin->audio_added)
      gst_element_remove_pad (GST_ELEMENT (ibin), ibin->audio_pad);
    else
      gst_object_unref (ibin->audio_pad);
  }
  if (ibin->subpicture_pad) {
    if (ibin->subpicture_added)
      gst_element_remove_pad (GST_ELEMENT (ibin), ibin->subpicture_pad);
    else
      gst_object_unref (ibin->subpicture_pad);
  }

  ibin->video_added = ibin->audio_added = ibin->subpicture_added = FALSE;
  ibin->audio_broken = FALSE;
  ibin->video_pad = ibin->audio_pad = ibin->subpicture_pad = NULL;
}

#if 0
static GstPad *
connect_thru_mq (GstInteractBin * ibin, GstPad * pad)
{
  GstPad *mq_sink;
  GstPad *mq_src;
  gchar *tmp, *sinkname, *srcname;

  /* Request a pad from multiqueue, then connect this one, then
   * discover the corresponding output pad and return it */
  mq_sink = gst_element_get_request_pad (ibin->pieces[GSTI_ELEM_MQUEUE],
      "sink%d");
  if (mq_sink == NULL)
    return FALSE;
  ibin->mq_req_pads = g_list_prepend (ibin->mq_req_pads, mq_sink);

  if (gst_pad_link (pad, mq_sink) != GST_PAD_LINK_OK)
    return FALSE;

  sinkname = gst_pad_get_name (mq_sink);
  tmp = sinkname + 4;
  srcname = g_strdup_printf ("src%s", tmp);

  mq_src = gst_element_get_static_pad (ibin->pieces[GSTI_ELEM_MQUEUE],
      srcname);

  g_free (sinkname);
  g_free (srcname);

  return mq_src;
}

static gboolean
can_sink_caps (GstElement * e, GstCaps * caps)
{
  gboolean res = FALSE;
  GstPad *sink = gst_element_get_static_pad (e, "sink");

  if (sink) {
    GstCaps *sink_caps = gst_pad_get_caps (sink);
    if (sink_caps) {
      res = gst_caps_can_intersect (sink_caps, caps);
      gst_caps_unref (sink_caps);
    }
    gst_object_unref (sink);
  }

  return res;
}
#endif

static void
ibin_pad_blocked_cb (GstPad * opad, gboolean blocked,
    GstInteractBinPadBlockCtx * ctx)
{
  GstInteractBin *ibin;
  GstPad *pad;
  gboolean added_last_pad = FALSE;
  gboolean added = FALSE;

  /* If not blocked ctx is NULL! */
  if (!blocked) {
    GST_DEBUG_OBJECT (opad, "Pad unblocked");
    return;
  }

  ibin = ctx->ibin;
  pad = ctx->pad;

  if (pad == ibin->subpicture_pad) {
    GST_DEBUG_OBJECT (opad, "Pad block -> subpicture pad");
    GST_IBIN_PREROLL_LOCK (ibin);
    added = ibin->subpicture_added;
    ibin->subpicture_added = TRUE;

    if (!added) {
      gst_element_add_pad (GST_ELEMENT (ibin), ibin->subpicture_pad);
      added_last_pad = ((ibin->audio_broken || ibin->audio_added)
          && ibin->video_added);
    }
    GST_IBIN_PREROLL_UNLOCK (ibin);

    gst_pad_set_blocked_async (opad, FALSE,
        (GstPadBlockCallback) ibin_pad_blocked_cb, NULL);
  } else if (pad == ibin->audio_pad) {
    GST_DEBUG_OBJECT (opad, "Pad block -> audio pad");
    GST_IBIN_PREROLL_LOCK (ibin);
    added = ibin->audio_added;
    ibin->audio_added = TRUE;

    if (!added) {
      gst_element_add_pad (GST_ELEMENT (ibin), ibin->audio_pad);
      added_last_pad = (ibin->subpicture_added && ibin->video_added);
    }
    GST_IBIN_PREROLL_UNLOCK (ibin);

    gst_pad_set_blocked_async (opad, FALSE,
        (GstPadBlockCallback) ibin_pad_blocked_cb, NULL);
  } else if (pad == ibin->video_pad) {
    GST_DEBUG_OBJECT (opad, "Pad block -> video pad");

    GST_IBIN_PREROLL_LOCK (ibin);
    added = ibin->video_added;
    ibin->video_added = TRUE;

    if (!added) {
      gst_element_add_pad (GST_ELEMENT (ibin), ibin->video_pad);
      added_last_pad = (ibin->subpicture_added && (ibin->audio_added
              || ibin->audio_broken));
    }
    GST_IBIN_PREROLL_UNLOCK (ibin);

    gst_pad_set_blocked_async (opad, FALSE,
        (GstPadBlockCallback) ibin_pad_blocked_cb, NULL);
  }

  if (added_last_pad) {
    GST_DEBUG_OBJECT (ibin, "Firing no more pads from pad-blocked cb");
    gst_element_no_more_pads (GST_ELEMENT (ibin));
  }
}

static void
gst_interactbin_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstInteractBin *ibin = GST_INTERACTBIN (object);

  switch (prop_id) {
    case ARG_LOCATION:
      GST_IBIN_LOCK (ibin);
      g_free (ibin->location);
      if (g_value_get_string (value) == NULL)
        ibin->location = g_strdup (DEFAULT_LOCATION);
      else
        ibin->location = g_value_dup_string (value);

      if (ibin->pieces[GSTI_ELEM_SOURCE]) {
        g_object_set_property (G_OBJECT (ibin->pieces[GSTI_ELEM_SOURCE]),
            "location", value);
      }
      GST_IBIN_UNLOCK (ibin);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_interactbin_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstInteractBin *ibin = GST_INTERACTBIN (object);

  switch (prop_id) {
    case ARG_LOCATION:
      GST_IBIN_LOCK (ibin);
      if (ibin->location)
        g_value_set_string (value, ibin->location);
      else if (ibin->pieces[GSTI_ELEM_SOURCE]) {
        g_object_get_property (G_OBJECT (ibin->pieces[GSTI_ELEM_SOURCE]),
            "location", value);
      } else
        g_value_set_string (value, DEFAULT_LOCATION);
      GST_IBIN_UNLOCK (ibin);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static GstStateChangeReturn
gst_interactbin_change_state (GstElement * element, GstStateChange transition)
{
  GstStateChangeReturn ret;
  GstInteractBin *ibin = GST_INTERACTBIN (element);

  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      if (!create_elements (ibin)) {
        remove_elements (ibin);
        return GST_STATE_CHANGE_FAILURE;
      }
      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

  switch (transition) {
    case GST_STATE_CHANGE_PAUSED_TO_READY:
    case GST_STATE_CHANGE_READY_TO_NULL:
      remove_elements (ibin);
      break;
    default:
      break;
  }

  return ret;
}
