/* GStreamer
 * Copyright (C) 2010 Jan Schmidt <thaytan@noraisin.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>

#include <gst/gst.h>
#include <gst/interfaces/navigation.h>

#include <cairo.h>

#include "gstinteractsrc.h"

GST_DEBUG_CATEGORY (gstinteractsrc_debug);
#define GST_CAT_DEFAULT gstinteractsrc_debug

#define DEFAULT_LOCATION ""

/* FIXME: Do negotiation on the source pad */
#define WIDTH  1024
#define HEIGHT 768
#define FPS    10

enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  ARG_0,
  ARG_LOCATION
};

static GstStaticPadTemplate src_factory =
  GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC, GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("video/x-raw-rgb,bpp=32,depth=32,blue_mask=-16777216,"
        "green_mask=16711680, red_mask=65280, alpha_mask=255,"
        "width=[ 1, 2147483647 ],height=[ 1, 2147483647 ],"
        "framerate=[ 0/1, 2147483647/1 ]")
  );

static GstStaticPadTemplate subpicture_src_template =
  GST_STATIC_PAD_TEMPLATE ("subpicture",
    GST_PAD_SRC, GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("video/x-dvd-subpicture")
  );

static void gst_interact_register_extra (GType gst_interact_type);

GST_BOILERPLATE_FULL (GstInteractSrc, gst_interact, GstBaseSrc,
    GST_TYPE_BASE_SRC, gst_interact_register_extra);

static void gst_interact_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_interact_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static void gst_interact_finalize (GObject * object);

static gboolean gst_interact_start (GstBaseSrc * bsrc);
static gboolean gst_interact_stop (GstBaseSrc * bsrc);
static gboolean gst_interact_unlock (GstBaseSrc * bsrc);
static gboolean gst_interact_unlock_stop (GstBaseSrc * bsrc);

static gboolean gst_interact_is_seekable (GstBaseSrc * bsrc);
static GstStateChangeReturn
gst_interact_change_state (GstElement * element, GstStateChange transition);

static gboolean gst_interact_check_get_range (GstBaseSrc * src);
static GstFlowReturn gst_interact_create (GstBaseSrc * bsrc, guint64 offset,
    guint length, GstBuffer ** buf);
static gboolean gst_interact_src_event (GstBaseSrc * basesrc, GstEvent * event);
static gboolean gst_interact_src_query (GstBaseSrc * basesrc, GstQuery * query);

static void send_segment_updates(GstInteractSrc *src,
  GstClockTime new_time, gboolean update);

static void
gst_interact_register_extra (GType gst_interact_type)
{
  GST_DEBUG_CATEGORY_INIT (gstinteractsrc_debug, "gstinteractsrc", 0,
      "Interactivity Demo source element");
}

static void
gst_interact_base_init (gpointer gclass)
{

  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&subpicture_src_template));

  gst_element_class_set_details_simple (element_class, "Interaction Demo Src",
      "Source/Example", "Interactivity Demo source element",
      "Jan Schmidt <thaytan@noraisin.net>");
}

static void
gst_interact_class_init (GstInteractSrcClass * klass)
{
  GObjectClass *gobject_class = (GObjectClass *) klass;
  GstElementClass *gstelement_class = (GstElementClass *) klass;
  GstBaseSrcClass *gstbasesrc_class = GST_BASE_SRC_CLASS (klass);

  gobject_class->finalize = gst_interact_finalize;
  gobject_class->set_property = gst_interact_set_property;
  gobject_class->get_property = gst_interact_get_property;

  gstelement_class->change_state = gst_interact_change_state;

  gstbasesrc_class->start = GST_DEBUG_FUNCPTR (gst_interact_start);
  gstbasesrc_class->stop = GST_DEBUG_FUNCPTR (gst_interact_stop);
  gstbasesrc_class->unlock = GST_DEBUG_FUNCPTR (gst_interact_unlock);
  gstbasesrc_class->unlock_stop = GST_DEBUG_FUNCPTR (gst_interact_unlock_stop);
  gstbasesrc_class->event = GST_DEBUG_FUNCPTR (gst_interact_src_event);
  gstbasesrc_class->query = GST_DEBUG_FUNCPTR (gst_interact_src_query);
  gstbasesrc_class->is_seekable = GST_DEBUG_FUNCPTR (gst_interact_is_seekable);

  gstbasesrc_class->check_get_range =
      GST_DEBUG_FUNCPTR (gst_interact_check_get_range);
  gstbasesrc_class->create = GST_DEBUG_FUNCPTR (gst_interact_create);

  g_object_class_install_property (gobject_class, ARG_LOCATION,
      g_param_spec_string ("location", "Location", "Slides location",
          NULL, G_PARAM_READWRITE));
}

static void
gst_interact_init (GstInteractSrc *src, GstInteractSrcClass * gclass)
{
  src->interact_lock = g_mutex_new();

  gst_base_src_set_format (GST_BASE_SRC (src), GST_FORMAT_TIME);

  src->spu_pad =
      gst_pad_new_from_static_template (&subpicture_src_template, "subpicture");

  gst_element_add_pad ((GstElement *) (src), src->spu_pad);

  src->slides = g_array_new (FALSE, FALSE, sizeof (GstInteractSlide));

  src->target_w = WIDTH;
  src->target_h = HEIGHT;

  /* Hack: should get the native slide width/height info from the first
   * slide. */
  src->full_w = 1440;
  src->full_h = 900;

  src->out_w = src->target_w;
  src->out_h = src->full_h * src->target_w / src->full_w;
}

static void
gst_interact_finalize (GObject * object)
{
  GstInteractSrc *src = GST_INTERACT_SRC (object);

  g_mutex_free (src->interact_lock);
  g_free (src->location);

  gst_buffer_replace (&src->out_buf, NULL);
  gst_buffer_replace (&src->last_buf, NULL);
  gst_caps_replace (&src->cur_caps, NULL);

  slides_free (src->slides);
  g_array_free (src->slides, TRUE);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

static gboolean
gst_interact_unlock (GstBaseSrc * bsrc)
{
  //GstInteractSrc *src = GST_INTERACT_SRC (bsrc);

  return TRUE;
}

static gboolean
gst_interact_unlock_stop (GstBaseSrc * bsrc)
{
  //GstInteractSrc *src = GST_INTERACT_SRC (bsrc);

  return TRUE;
}

static void
gst_interact_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstInteractSrc *src = GST_INTERACT_SRC (object);

  switch (prop_id) {
    case ARG_LOCATION:
      GST_OBJECT_LOCK (src);
      g_free (src->location);
      if (g_value_get_string (value) == NULL)
        src->location = g_strdup (DEFAULT_LOCATION);
      else
        src->location = g_value_dup_string (value);
      GST_OBJECT_UNLOCK (src);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_interact_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstInteractSrc *src = GST_INTERACT_SRC (object);

  switch (prop_id) {
    case ARG_LOCATION:
      GST_OBJECT_LOCK (src);
      g_value_set_string (value, src->location);
      GST_OBJECT_UNLOCK (src);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
change_slide (GstInteractSrc *src, GstInteractSlide *slide)
{
  src->prev_slide = src->cur_slide;
  src->cur_slide = slide;
}

static gboolean
gst_interact_start (GstBaseSrc * bsrc)
{
  GstInteractSrc *src = GST_INTERACT_SRC (bsrc);

  src->running = TRUE;
  src->need_redraw = TRUE;
  src->last_x = src->last_y = -1;
  src->commands_changed = TRUE;

  if (!slides_read(src, src->location)) {
    GST_ELEMENT_ERROR (src, CORE, FAILED, (NULL),
      ("Failed to read configuration file"));
    return FALSE;
  }
  if (src->slides->len == 0) {
    GST_ELEMENT_ERROR (src, CORE, FAILED, (NULL),
      ("No slides in configuration file"));
    return FALSE;
  }

  change_slide (src, &g_array_index (src->slides, GstInteractSlide, 0));
  src->frame_no = 0;

  send_segment_updates (src, 0, FALSE);

  return TRUE;
}

static gboolean
gst_interact_stop (GstBaseSrc * bsrc)
{
  GstInteractSrc *src = GST_INTERACT_SRC (bsrc);
  gboolean ret = TRUE;
  GstMessage *mouse_over_msg = NULL;

  src->running = TRUE;

  if (src->was_mouse_over) {
    mouse_over_msg =
        gst_navigation_message_new_mouse_over ((GstObject *) src, FALSE);
    src->was_mouse_over = FALSE;
  }

  if (mouse_over_msg)
    gst_element_post_message (GST_ELEMENT_CAST (src), mouse_over_msg);

  slides_free (src->slides);

  return ret;
}

/* Send app a bus message that the available commands have changed */
static void
gst_interact_send_commands_changed (GstInteractSrc * src)
{
  GstMessage *cmds_msg =
      gst_navigation_message_new_commands_changed (GST_OBJECT_CAST (src));
  gst_element_post_message (GST_ELEMENT_CAST (src), cmds_msg);
}

static gboolean
gst_interact_handle_cmds_query (GstInteractSrc * src, GstQuery * query)
{
  /* Expand this array if we have more commands in the future: */
  GstNavigationCommand cmds[6];
  gint n_cmds = 0;

  /* Fill out the standard set of commands we support */
  cmds[n_cmds++] = GST_NAVIGATION_COMMAND_DVD_MENU;

  g_mutex_lock (src->interact_lock);

  /* Add button selection commands if we have them */
  if (src->active_button > 0) {
    /* We have a valid current button */
    cmds[n_cmds++] = GST_NAVIGATION_COMMAND_ACTIVATE;
  }
#if 0
  /* Check for buttons in each direction */
  if (src->cur_btn_mask & RSN_BTN_LEFT)
    cmds[n_cmds++] = GST_NAVIGATION_COMMAND_LEFT;
  if (src->cur_btn_mask & RSN_BTN_RIGHT)
    cmds[n_cmds++] = GST_NAVIGATION_COMMAND_RIGHT;
  if (src->cur_btn_mask & RSN_BTN_UP)
    cmds[n_cmds++] = GST_NAVIGATION_COMMAND_UP;
  if (src->cur_btn_mask & RSN_BTN_DOWN)
    cmds[n_cmds++] = GST_NAVIGATION_COMMAND_DOWN;
#endif
  g_mutex_unlock (src->interact_lock);

  gst_navigation_query_set_commandsv (query, n_cmds, cmds);

  return TRUE;
}

static gboolean
gst_interact_handle_navigation_query (GstInteractSrc * src,
    GstNavigationQueryType nq_type, GstQuery * query)
{
  gboolean res;

  GST_LOG_OBJECT (src, "Have Navigation query of type %d", nq_type);

  switch (nq_type) {
    case GST_NAVIGATION_QUERY_COMMANDS:
      res = gst_interact_handle_cmds_query (src, query);
      break;
    default:
      res = FALSE;
  }

  return res;
}

static GstCaps *
make_rgba_caps(gint w, gint h)
{
  GstCaps *c;
  gchar *s = g_strdup_printf("video/x-raw-rgb,endianness=4321,bpp=32,"
                "depth=32,blue_mask=-16777216,green_mask=16711680,"
                "red_mask=65280,alpha_mask=255,width=%d,height=%d,"
                "framerate=%d/1", w, h, FPS);
  c = gst_caps_from_string (s); 
  g_free (s);

  return c;
}

static void
draw_stamp (GstInteractSrc *src, cairo_t *cr, GstClockTime ts)
{
  cairo_text_extents_t te;
  gchar *s;
  gint xoff, yoff;
  xoff = (src->target_w - src->out_w) / 2;
  yoff = (src->target_h - src->out_h) / 2;

  s = g_strdup_printf("%02d:%02d:%02d",
    (gint)(ts / 3600), (gint)((ts / 60) % 60), (gint)(ts % 60));

  cairo_set_source_rgba (cr, 1.0, 1.0, 1.0, 1.0);
  cairo_select_font_face (cr, "Monospace",
    CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_BOLD);
  cairo_set_font_size (cr, 24);
  cairo_text_extents (cr, s, &te);
  cairo_translate (cr, -xoff, -yoff);
  cairo_move_to (cr, src->target_w - te.width - te.x_bearing -
    (20 * src->out_w / src->full_w),
    src->target_h - 2 * (te.height + te.y_bearing) -
    (60 * src->out_h / src->full_h));
  cairo_show_text (cr, s);
  g_free (s);

#if 0
  if (src->last_x != -1) {
    s = g_strdup_printf("(%d, %d)",
            (src->last_x - xoff) * src->full_w / src->out_w,
            (src->last_y - yoff) * src->full_h / src->out_h);
    cairo_text_extents (cr, s, &te);
    cairo_move_to (cr, src->target_w - te.width - te.x_bearing -
      (20 * src->out_w / src->full_w),
      src->target_h - (te.height + te.y_bearing) -
      (10 * src->out_h / src->full_h));
    cairo_show_text (cr, s);
    g_free (s);
  }
#endif
}

static GstFlowReturn
gst_interact_prepare_next_buf (GstInteractSrc * src)
{
  GstFlowReturn ret = GST_FLOW_OK;
  GstBuffer *buf;
  GstClockTime ts;

  src->out_buf = buf = gst_buffer_new_and_alloc (4 * src->target_w * src->target_h);
  if (src->cur_caps == NULL)
    src->cur_caps = make_rgba_caps (src->target_w, src->target_h);

  gst_buffer_set_caps (buf, src->cur_caps);

  ts = gst_util_uint64_scale_int (GST_SECOND, src->frame_no, FPS);
  GST_BUFFER_TIMESTAMP (buf) = ts;
  src->frame_no++;

  /* Check if the timestamp changed, before redrawing */
  if (src->last_time != (ts / GST_SECOND)) {
    src->need_redraw = TRUE;
    src->last_time = ts / GST_SECOND;
  }
  if (src->last_buf == NULL)
    src->need_redraw = TRUE;

  if (src->need_redraw) {
    cairo_surface_t *surface;
    cairo_t *ctx;
    RsvgHandle *rsvg;

    surface = cairo_image_surface_create_for_data (
                 GST_BUFFER_DATA (buf), CAIRO_FORMAT_ARGB32,
                 src->target_w, src->target_h, src->target_w * 4);
  
    ctx = cairo_create (surface);
  
    rsvg = src->cur_slide->rsvg;
    if (rsvg) {
       gint xoff, yoff;
       xoff = (src->target_w - src->out_w) / 2;
       yoff = (src->target_h - src->out_h) / 2;

       cairo_translate (ctx, xoff, yoff);
       rsvg_handle_render_cairo (rsvg, ctx);
       cairo_translate (ctx, -xoff, -yoff);
    }

    draw_stamp (src, ctx, src->last_time);
    cairo_destroy (ctx);
    cairo_surface_finish (surface);

    src->need_redraw = FALSE;
  }
  else {
    memcpy (GST_BUFFER_DATA (src->out_buf), GST_BUFFER_DATA (src->last_buf),
      GST_BUFFER_SIZE (src->out_buf));
  }

  return ret;
}

static gboolean
gst_interact_check_get_range (GstBaseSrc * src)
{
  return FALSE;
}

static void
send_segment_updates(GstInteractSrc *src, GstClockTime new_time, gboolean update)
{
  GstSegment *segment = &(GST_BASE_SRC (src)->segment);
  GstEvent *event;

  event = gst_event_new_new_segment_full (update,
              segment->rate, segment->applied_rate,
              GST_FORMAT_TIME, new_time,
              segment->stop,
              segment->time + (new_time - segment->start));
  gst_pad_push_event (src->spu_pad, event);
}

static GstFlowReturn
gst_interact_create (GstBaseSrc * bsrc, guint64 offset,
    guint length, GstBuffer ** outbuf)
{
  GstInteractSrc *src = GST_INTERACT_SRC (bsrc);
  GstFlowReturn ret;
  gboolean cmds_changed = FALSE;

  *outbuf = NULL;

  g_mutex_lock (src->interact_lock);
  ret = gst_interact_prepare_next_buf (src);
  if (ret != GST_FLOW_OK) {
    g_mutex_unlock (src->interact_lock);
    return ret;
  }

  cmds_changed = src->commands_changed;
  src->commands_changed = FALSE;

  if (src->out_buf != NULL) {
    gst_buffer_replace (&src->last_buf, src->out_buf);

    *outbuf = src->out_buf;
    src->out_buf = NULL;

    if (src->discont) {
      GST_LOG_OBJECT (src, "Marking discont buffer");
      GST_BUFFER_FLAG_SET (*outbuf, GST_BUFFER_FLAG_DISCONT);
      src->discont = FALSE;
    }
  }

  g_mutex_unlock (src->interact_lock);

  if (cmds_changed)
    gst_interact_send_commands_changed (src);

  if (*outbuf)
    send_segment_updates(src, GST_BUFFER_TIMESTAMP (*outbuf), TRUE);

  return ret;
}

static gboolean
gst_interact_perform_button_action (GstInteractSrc * src,
    GstNavigationCommand action)
{
  gboolean result = FALSE;

  switch (action) {
    case GST_NAVIGATION_COMMAND_ACTIVATE:
      break;
    case GST_NAVIGATION_COMMAND_LEFT:
      break;
    case GST_NAVIGATION_COMMAND_RIGHT:
      break;
    case GST_NAVIGATION_COMMAND_DOWN:
      break;
    case GST_NAVIGATION_COMMAND_UP:
      break;
    default:
      break;
  }

  return result;
}

static gboolean
gst_interact_do_command (GstInteractSrc * src, GstNavigationCommand command)
{
  gboolean result = FALSE;

  switch (command) {
    case GST_NAVIGATION_COMMAND_DVD_MENU:
      break;
    case GST_NAVIGATION_COMMAND_LEFT:
    case GST_NAVIGATION_COMMAND_RIGHT:
    case GST_NAVIGATION_COMMAND_UP:
    case GST_NAVIGATION_COMMAND_DOWN:
    case GST_NAVIGATION_COMMAND_ACTIVATE:
      return gst_interact_perform_button_action (src, command);
    default:
      break;
  }

  return result;
}

static void
spawn_cmd (GstInteractSrc *src, const gchar *cmd)
{
  GError *error = NULL;
  g_spawn_command_line_async (cmd, &error);
  if (error) {
    GST_ELEMENT_ERROR (src, RESOURCE, FAILED,
        ("Unable to spawn command: error %s\n", error->message),
        (NULL));
    g_error_free (error);
  }
}

static gboolean
gst_interact_handle_navigation_event (GstInteractSrc * src, GstEvent * event)
{
  gboolean have_lock = FALSE;
  gboolean nav_res = FALSE;
  gboolean channel_hop = FALSE;
  GstNavigationEventType etype = gst_navigation_event_get_type (event);
  GstMessage *mouse_over_msg = NULL;

  switch (etype) {
    case GST_NAVIGATION_EVENT_KEY_PRESS:{
      const gchar *key;
      if (!gst_navigation_event_parse_key_event (event, &key))
        return FALSE;

      GST_DEBUG_OBJECT (src, "got a keypress: %s", key);

      g_mutex_lock (src->interact_lock);
      have_lock = TRUE;
      if (!src->running)
        goto not_running;

      if (g_str_equal (key, "Return")) {
        nav_res = gst_interact_do_command (src, GST_NAVIGATION_COMMAND_ACTIVATE);
      } else if (g_str_equal (key, "Left")) {
        nav_res = gst_interact_do_command (src, GST_NAVIGATION_COMMAND_LEFT);
      } else if (g_str_equal (key, "Right")) {
        nav_res = gst_interact_do_command (src, GST_NAVIGATION_COMMAND_RIGHT);
      } else if (g_str_equal (key, "Up")) {
        nav_res = gst_interact_do_command (src, GST_NAVIGATION_COMMAND_UP);
      } else if (g_str_equal (key, "Down")) {
        nav_res = gst_interact_do_command (src, GST_NAVIGATION_COMMAND_DOWN);
      } else if (g_str_equal (key, "m")) {
        nav_res = gst_interact_do_command (src, GST_NAVIGATION_COMMAND_DVD_MENU);
      } 
      break;
    }
    case GST_NAVIGATION_EVENT_MOUSE_MOVE:{
      gdouble x, y;
      GstInteractButton *button;

      if (!gst_navigation_event_parse_mouse_move_event (event, &x, &y))
        return FALSE;

      button = slides_find_button (src, src->cur_slide, x, y);
      if (button) {
        if (!src->was_mouse_over) {
          GST_DEBUG_OBJECT (src, "Mouse moved onto a button");
          mouse_over_msg =
              gst_navigation_message_new_mouse_over ((GstObject *) src, TRUE);
          src->was_mouse_over = TRUE;
        }
      }
      else {
        if (src->was_mouse_over) {
          GST_DEBUG_OBJECT (src, "Mouse moved out of a button");
          mouse_over_msg =
              gst_navigation_message_new_mouse_over ((GstObject *) src, FALSE);
          src->was_mouse_over = FALSE;
        }
      }

      g_mutex_lock (src->interact_lock);
      have_lock = TRUE;
      if (!src->running)
        goto not_running;

      break;
    }
    case GST_NAVIGATION_EVENT_MOUSE_BUTTON_RELEASE:{
      gdouble x, y;
      gint mouse_button;
      GstInteractButton *button;

      if (!gst_navigation_event_parse_mouse_button_event (event, &mouse_button,
              &x, &y))
        return FALSE;
      if (mouse_button != 1)
        return FALSE;

      GST_DEBUG_OBJECT (src, "Got click at %g, %g", x, y);

      g_mutex_lock (src->interact_lock);
      have_lock = TRUE;
      if (!src->running)
        goto not_running;
      src->last_x = x;
      src->last_y = y;
      src->need_redraw = TRUE;
      button = slides_find_button (src, src->cur_slide, x, y);
      if (button) {
        GstInteractSlide *slide;
        if (g_str_equal (button->dest_name, "BACK"))
          slide = src->prev_slide;
        else if (g_str_equal (button->dest_name, "LAUNCH")) {
          spawn_cmd (src, button->launch_cmd);
        } else
          slide = slides_find_slide (src, button->dest_name);
        if (slide) {
          change_slide (src, slide);
          channel_hop = TRUE;
        }
      }
      break;
    }
    case GST_NAVIGATION_EVENT_COMMAND:{
      GstNavigationCommand command;

      if (!gst_navigation_event_parse_command (event, &command))
        return FALSE;
      if (command == GST_NAVIGATION_COMMAND_INVALID)
        return FALSE;

      g_mutex_lock (src->interact_lock);
      have_lock = TRUE;
      if (!src->running)
        goto not_running;

      GST_LOG_OBJECT (src, "handling navigation command %d", command);
      nav_res = gst_interact_do_command (src, command);
      break;
    }
    default:
      return TRUE;
  }

  if (have_lock) {
    gboolean cmds_changed;

    if (channel_hop) {
      src->need_redraw = TRUE;
      src->commands_changed = TRUE;
    }

    cmds_changed = src->commands_changed;
    src->commands_changed = FALSE;

    g_mutex_unlock (src->interact_lock);

    if (channel_hop) {
      GstEvent *seek;
      if (src->was_mouse_over) {
          GST_DEBUG_OBJECT (src, "Clearing mouse over due to flush");
          if (mouse_over_msg)
            gst_message_unref (mouse_over_msg);
          mouse_over_msg =
              gst_navigation_message_new_mouse_over ((GstObject *) src, FALSE);
          src->was_mouse_over = FALSE;
      }

      /* Send ourselves a seek event to wake everything up and flush */
      seek = gst_event_new_seek (1.0, GST_FORMAT_TIME, GST_SEEK_FLAG_FLUSH,
          GST_SEEK_TYPE_NONE, 0, GST_SEEK_TYPE_NONE, -1);
      gst_element_send_event (GST_ELEMENT (src), seek);
    }

    if (cmds_changed)
      gst_interact_send_commands_changed (src);
  }

  if (mouse_over_msg) {
    gst_element_post_message (GST_ELEMENT_CAST (src), mouse_over_msg);
  }

  return TRUE;
not_running:
  if (have_lock)
    g_mutex_unlock (src->interact_lock);
  GST_DEBUG_OBJECT (src, "Element not started. Ignoring navigation event");
  return FALSE;
}

static gboolean
gst_interact_src_event (GstBaseSrc * basesrc, GstEvent * event)
{
  GstInteractSrc *src = GST_INTERACT_SRC (basesrc);
  gboolean res;

  switch (GST_EVENT_TYPE (event)) {
    case GST_EVENT_NAVIGATION:
      res = gst_interact_handle_navigation_event (src, event);
      break;
    case GST_EVENT_SEEK:{
      res = GST_BASE_SRC_CLASS (parent_class)->event (basesrc, event);
      break;
    }
    default:
      GST_LOG_OBJECT (src, "handling %s event", GST_EVENT_TYPE_NAME (event));

      res = GST_BASE_SRC_CLASS (parent_class)->event (basesrc, event);
      break;
  }

  return res;
}

static GstStateChangeReturn
gst_interact_change_state (GstElement * element, GstStateChange transition)
{
  GstStateChangeReturn ret;
  GstInteractSrc *src = GST_INTERACT_SRC (element);

  switch (transition) {
    case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
      GST_DEBUG_OBJECT (element, "Switching to PAUSED");
      g_mutex_lock (src->interact_lock);
      src->in_playing = FALSE;
      g_mutex_unlock (src->interact_lock);
      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

  switch (transition) {
    case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
      GST_DEBUG_OBJECT (element, "Switching to PLAYING");
      g_mutex_lock (src->interact_lock);
      src->in_playing = TRUE;
      src->commands_changed = TRUE;
      g_mutex_unlock (src->interact_lock);
      break;
    default:
      break;
  }

  return ret;
}

static gboolean
gst_interact_src_query (GstBaseSrc * basesrc, GstQuery * query)
{
  GstInteractSrc *src = GST_INTERACT_SRC (basesrc);
  gboolean res = FALSE;
  GstFormat format;

  switch (GST_QUERY_TYPE (query)) {
    case GST_QUERY_DURATION:
      break;
    case GST_QUERY_POSITION:
      gst_query_parse_position (query, &format, NULL);
      break;
    case GST_QUERY_CUSTOM:
    {
      GstNavigationQueryType nq_type = gst_navigation_query_get_type (query);
      if (nq_type != GST_NAVIGATION_QUERY_INVALID)
        res = gst_interact_handle_navigation_query (src, nq_type, query);
      else
        res = GST_BASE_SRC_CLASS (parent_class)->query (basesrc, query);
      break;
    }
    default:
      res = GST_BASE_SRC_CLASS (parent_class)->query (basesrc, query);
      break;
  }

  return res;
}

static gboolean
gst_interact_is_seekable (GstBaseSrc * bsrc)
{
  return FALSE;
}

