/* GStreamer
 * Copyright (C) 2010 Jan Schmidt <thaytan@noraisin.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>

#include <gst/gst.h>

#include "gstinteractsrc.h"

GST_DEBUG_CATEGORY_EXTERN (gstinteractsrc_debug);
#define GST_CAT_DEFAULT gstinteractsrc_debug

struct SlideCtx {
  GstInteractSlide *cur_slide;
  gchar *base_path;
};

static void
handle_svg_sizing (gint *width, gint *height, gpointer user_data)
{
  GstInteractSrc *src = GST_INTERACT_SRC (user_data);

  src->full_w = *width;
  src->full_h = *height;

  src->out_w = *width = src->target_w;
  src->out_h = *height = src->full_h * src->target_w / src->full_w;
}

static gboolean
process_line (GstInteractSrc *src, struct SlideCtx *ctx, gchar *in, gchar *end)
{
  gboolean res = FALSE;
  gchar *line = g_strndup (in, end - in);
  gchar **parts = g_strsplit (line, " ", 0);
  gint n_parts = g_strv_length (parts);

  GST_LOG_OBJECT (src, "handle config line %s", line);
  if (g_str_has_prefix (line, "SLIDE")) {
    GstInteractSlide new_slide;
    gchar *fname;

    memset (&new_slide, 0, sizeof (GstInteractSlide));
    if (n_parts < 3) {
      GST_DEBUG ("Invalid SLIDE line: %s", line);
      goto out;
    }

    fname = g_build_filename (ctx->base_path, parts[2], NULL);
    new_slide.rsvg = rsvg_handle_new_from_file (fname, NULL);
    if (new_slide.rsvg == NULL) {
      GST_DEBUG ("Failed to read SVG for SLIDE line: %s", line);
      g_free (fname);
      goto out;
    }
    rsvg_handle_set_size_callback (new_slide.rsvg, handle_svg_sizing,
        src, NULL);

    new_slide.name = g_strdup (parts[1]);
    new_slide.buttons = g_array_new(FALSE, FALSE, sizeof (GstInteractButton));

    g_array_append_val (src->slides, new_slide);
    ctx->cur_slide =
       &g_array_index (src->slides, GstInteractSlide, src->slides->len-1);
  }
  else if (g_str_has_prefix (line, "BUTTON")) {
    GstInteractButton new_button;
    memset (&new_button, 0, sizeof (GstInteractButton));
    if (ctx->cur_slide == NULL) {
      GST_DEBUG ("No current slide for BUTTON: %s", line);
      goto out;
    }
    if (n_parts < 7) {
      GST_DEBUG ("Invalid BUTTON: %s", line);
      goto out;
    }
    new_button.id = g_ascii_strtoll (parts[1], NULL, 10);
    new_button.x1 = g_ascii_strtoll (parts[2], NULL, 10);
    new_button.y1 = g_ascii_strtoll (parts[3], NULL, 10);
    new_button.x2 = g_ascii_strtoll (parts[4], NULL, 10);
    new_button.y2 = g_ascii_strtoll (parts[5], NULL, 10);
    new_button.dest_name = g_strdup (parts[6]);
    if (g_str_equal (new_button.dest_name, "LAUNCH")) {
      new_button.launch_cmd = g_strjoinv (" ", parts + 7); 
      if (n_parts < 8) {
        GST_DEBUG ("Invalid BUTTON: %s", line);
        goto out;
      }
    }    

    g_array_append_val (ctx->cur_slide->buttons, new_button);
  }

  res = TRUE;
out:
  g_strfreev (parts);
  g_free (line);
  return res;
}

gboolean
slides_read(GstInteractSrc *src, const char *fname)
{
  gboolean res = FALSE;
  gsize len;
  gchar *c, *cur, *end;
  struct SlideCtx ctx = { 0, };
  int line = 0;

  if (!g_file_get_contents (fname, &c, &len, NULL)) {
    GST_ERROR_OBJECT (src, "Failed to read config %s", fname);
    return FALSE;
  }
  end = c + len;

  ctx.base_path = g_path_get_dirname (src->location);

  cur = c;
  while (cur != NULL) {
    gchar *next = strstr(cur, "\n");
    if (next == NULL) {
      line++;
      if (!process_line (src, &ctx, cur, end))
        goto out;
      break;
    }
    line++;
    if (!process_line (src, &ctx, cur, next))
      goto out;
    cur = next + 1;
  }

  res = TRUE;
out:
  g_free (ctx.base_path);
  if (!res) {
    GST_ELEMENT_ERROR (src, CORE, FAILED,
       ("Failed to read slides configuration in %s line %d\n", fname, line),
       (NULL));
  }
  g_free (c);
  return res;
}

static void
buttons_free (GArray *buttons)
{
  gint i;
  for (i = 0; i < buttons->len; i++) {
    GstInteractButton *b = &g_array_index (buttons, GstInteractButton, i);
    g_free (b->dest_name);
    g_free (b->launch_cmd);
  }
  g_array_set_size (buttons, 0);
}

void
slides_free (GArray *slides)
{
  gint i;
  for (i = 0; i < slides->len; i++) {
    GstInteractSlide *s = &g_array_index (slides, GstInteractSlide, i);
    g_free (s->name); 
    rsvg_handle_free (s->rsvg);
    buttons_free (s->buttons);
    g_array_free (s->buttons, TRUE);
  }
  g_array_set_size (slides, 0);
}

GstInteractButton *
slides_find_button (GstInteractSrc *src,
   GstInteractSlide *cur_slide, gint x, gint y)
{
  gint b;
  gint xoff, yoff;
  xoff = (src->target_w - src->out_w) / 2;
  yoff = (src->target_h - src->out_h) / 2;

  x -= xoff;
  y -= yoff;

  x = (src->full_w * x) / src->out_w;
  y = (src->full_h * y) / src->out_h;

  for (b = 0; b < cur_slide->buttons->len; b++) {
    GstInteractButton *cur_button =
        &g_array_index (cur_slide->buttons, GstInteractButton, b);
    if (x >= cur_button->x1 && x <= cur_button->x2 &&
        y >= cur_button->y1 && y <= cur_button->y2) {
      GST_INFO_OBJECT (src, "Got button press to slide %s",
          cur_button->dest_name);     
      return cur_button;
    }
  }

  return NULL;
}

GstInteractSlide *
slides_find_slide (GstInteractSrc *src, const gchar *name)
{
  gint s;
  for (s = 0; s < src->slides->len; s++) {
    GstInteractSlide *cur_slide =
        &g_array_index (src->slides, GstInteractSlide, s);
    if (g_str_equal (cur_slide->name, name))
        return cur_slide;
  }

  return NULL;
}
